/**
 * Module dependencies.
 */

var Suite = require('mocha/lib/suite');
var Test = require('mocha/lib/test');
var event = require('../event');
var scenario = require('../scenario');
var escapeRe = require('escape-string-regexp');
var utils = require('./utils');
var csv = require('csv-parser');
var fs = require('fs');
var path = require('path');

/**
 * Codecept-style interface:
 *
 * Feature('login')
 * Scenario('login as regular user', (I) {
 *   I.fillField();
 *   I.click()
 *   I.see('Hello, '+data.login);
 * });
 *
 * @param {Suite} suite Root suite.
 */
module.exports = function (suite) {
  var suites = [suite];

  suite.on('pre-require', function (context, file, mocha) {
    var common = require('mocha/lib/interfaces/common')(suites, context);

    common.before('codeceptjs event', function () {
      event.dispatcher.emit(event.suite.before);
    });

    common.after('codeceptjs event', function () {
      event.dispatcher.emit(event.suite.after);
    });
    

    // create dispatcher
    context.BeforeAll = common.before;
    context.AfterAll = common.after;

    context.run = mocha.options.delay && common.runWithSuite(suite);
    /**
     * Describe a "suite" with the given `title`
     * and callback `fn` containing nested suites
     * and/or tests.
     */

    context.Feature = function (title) { 
      if (suites.length > 1) {
        suites.shift();
      }         
      var suite = Suite.create(suites[0], title);
      suite.file = file;
      suite.timeout(0);
      suites.unshift(suite);
      suite.beforeEach('codeceptjs.before', scenario.setup);
      suite.afterEach('finialize codeceptjs', scenario.teardown);     
      return suite;
    };

    context.Background = context.Before = context.BeforeAll = function (fn) {
      suites[0].beforeEach('Before', scenario.injected(fn));
    };

    context.After = function (fn) {
      suites[0].afterEach('After', scenario.injected(fn));
    };

    /**
     * Describe a specification or test-case
     * with the given `title` and callback `fn`
     * acting as a thunk.
     */
    context.Scenario = function (title, fn) {
      var suite = suites[0];
      if (suite.pending) {
        fn = null;
      }
      var test = new Test(title, fn);
      test.file = file;
      test.async = true;
      test.timeout(0);
      suite.addTest(scenario.test(test));
      return test;
    };

    context.Scenario.OutlineFromCsv = function(title, fn, file) {
      var fileFullPath = path.join(global.data_dir, file);
      data = fs.readFileSync(fileFullPath);
      var array = data.toString().split("\n");
      var dataList = createAnObjectFromArray(array);
      context.Scenario.Outline(title, fn, dataList);
    };

    context.Scenario.Outline = function (title, fn, dataList) {
      //checks argument list is valid
      if(arguments.length !== 3) {
        throw new Error('Not enough arguments for Outline function. Arguments must be:' + 
          'title: scenario outline name' + 
          'fn: scenario I function' + 
          'dataList: each sceanrio"s data');
      }
      if ( !Array.isArray(dataList) ) {
        throw new Error('Outline needs an parameter object array');
      }

      // create new scenarios according to dataList array values
      for(var i = 0; i < dataList.length; i++) {
        // changes function code appropriate {replacable} arguments value 
        var newFn = utils.replace(fn, dataList[i]);
        //adds new scenario for each outline data object
        context.Scenario(title + "[Outline_" + i + "]", newFn);
      }
    };

    /**
     * Exclusive test-case.
     */
    context.Scenario.only = function (title, fn) {
      var test = context.Scenario(title, fn);
      var reString = '^' + escapeRe(test.fullTitle()) + '$';
      mocha.grep(new RegExp(reString));
    };

    /**
     * Pending test case.
     */
    context.xScenario = context.Scenario.skip = function (title) {
      context.Scenario(title);
    };

    /**
     * Pending outline test cases.
     */
    context.xScenario.Outline = context.Scenario.Outline.skip = function (title) {
      context.Scenario.skip(title + " [Outline]");
    }

    /**
     * Pending outline-json test cases.
     */
    context.xScenario.OutlineFromCsv = context.Scenario.OutlineFromCsv.skip = function (title) {
      context.Scenario.skip(title + " [Outline - with data]");
    }
  });
};

function createAnObjectFromArray(arr) {
  var result = [];
  var headers = arr[0].split(",");
  for(i = 1; i < arr.length; i++) {
    var tempObj = {};
    var datas = arr[i].split(",");
    for (j = 0; j<headers.length; j++) {
      tempObj[headers[j]] = datas[j];
    }
    result.push(tempObj);
  }
  return result;
};
